/*
 * Copyright 2000 Hans Reiser
 */
#include "fsck.h"



#if 0
/* find stat datas which were not reached during semantic pass and link them
   to "lost+found" */
static int link_lost (reiserfs_filsys_t fs, struct buffer_head ** path, int h)
{
    int i;
    struct item_head * ih;
    struct buffer_head * bh = path[h];
    char lost_name[80];

    ih = B_N_PITEM_HEAD (bh, 0);
    for (i = 0; i < B_NR_ITEMS (bh); i ++, ih++) {
	if (is_stat_data_ih (ih) && !is_item_reachable (ih)) {
	    struct key key = {0, 0, {{0, 0},}};

	    sprintf (lost_name, "%u_%u", le32_to_cpu (ih->ih_key.k_dir_id),
		     le32_to_cpu (ih->ih_key.k_objectid));
	    /* entry in lost+found directory will point to this key */
	    key.k_dir_id = ih->ih_key.k_dir_id;
	    key.k_objectid = ih->ih_key.k_objectid;

	    /* 0 does not mean anyting - item w/ "." and ".." already
               exists and reached, so only name will be added */
	    reiserfs_add_entry (fs, &lost_found_dir_key, lost_name, &key, 0, 1/*lost&found*/);
	    stat_lost_found (fs);
	}
    }
    return 0;
}


void pass_3a_look_for_lost (reiserfs_filsys_t s)
{
    if (!lost_found_dir_key.k_objectid) {
	fsck_progress ("look_for_lost: could not create lost+found directory\n");
	return;
    }

    /* pass through the tree and link stat data which were not reached
       yet to /lost+found */
    fsck_progress ("Looking for lost files..");
    pass_through_tree (s, 0, link_lost);
    fsck_progress ("ok\n");

    /* check names added we just have added to/lost+found. Those names
       are marked DEH_Lost_found flag */
    fsck_progress ("Checking lost+found directory.."); fflush (stdout);
    check_semantic_tree (&lost_found_dir_key, &root_dir_key, 0, 1/* lost+found*/);
    fsck_progress ("ok\n");

    stage_report (0x3a, fs);
}
#endif


/* fixme: search_by_key is not needed after any add_entry */
static void _look_for_lost (reiserfs_filsys_t fs, int dir)
{
    struct key key, prev_key;
    struct path path;
    int i;
    struct buffer_head * bh;
    struct item_head * ih;
    unsigned long leaves;

    path.path_length = ILLEGAL_PATH_ELEMENT_OFFSET;
    key = root_dir_key;
    
    fsck_progress ("Looking for lost %s.. - ", dir ? "directories" : "files");
    leaves = 0;

    while (usearch_by_key (fs, &key, &path) == ITEM_FOUND) {

	/* print ~ how many leaves were scanned and how fast it was */
	print_how_fast (0, &leaves, 50);

	bh = PATH_PLAST_BUFFER (&path);
	for (i = get_item_pos (&path), ih = get_ih (&path); i < B_NR_ITEMS (bh); i ++, ih ++) {
	    if (is_item_reachable (ih) || !is_stat_data_ih (ih))
		continue;
	    if (dir && not_a_directory (B_I_PITEM (bh,ih)))
		continue;
	  
	    {
		struct key obj_key = {0, 0, {{0, 0},}};
		char lost_name[80];

		sprintf (lost_name, "%u_%u", le32_to_cpu (ih->ih_key.k_dir_id),
			 le32_to_cpu (ih->ih_key.k_objectid));
		/* entry in lost+found directory will point to this key */
		obj_key.k_dir_id = ih->ih_key.k_dir_id;
		obj_key.k_objectid = ih->ih_key.k_objectid;

		/* need this to continue */
		prev_key = key;
		get_next_key (&path, i, &key);
		if (comp_keys (&prev_key, &key) != -1)
		    reiserfs_panic ("pass_3a: key must grow: prev=%k next=%k",
				    &prev_key, &key);
		pathrelse (&path);
		
		/* 0 does not mean anyting - item w/ "." and ".." already
		   exists and reached, so only name will be added */
		reiserfs_add_entry (fs, &lost_found_dir_key, lost_name, &obj_key, 0, 1/*lost&found*/);
		stat_lost_found (fs);

		goto cont;
	    }
	}
	prev_key = key;
	get_next_key (&path, i - 1, &key);
	if (comp_keys (&prev_key, &key) != -1)
	    reiserfs_panic ("pass_3a: key must grow 2: prev=%k next=%k",
			    &prev_key, &key);
	pathrelse (&path);

    cont:
    }

    pathrelse (&path);
    fsck_progress ("ok\n");

    /* check names added we just have added to/lost+found. Those names are
       marked DEH_Lost_found flag */
    fsck_progress ("Checking lost+found directory.."); fflush (stdout);
    check_semantic_tree (&lost_found_dir_key, &root_dir_key, 0, 1/* lost+found*/);
    fsck_progress ("ok\n");

}


void pass_3a_look_for_lost (reiserfs_filsys_t fs)
{
    /* look for lost dirs first */
    _look_for_lost (fs, 1);

    /* link files which are still lost */
    _look_for_lost (fs, 0);

    stage_report (0x3a, fs);
}

